/**	\file ComplexType.cpp
 **	\date  2008-05-26
 **	\author grymse@alhem.net
**/
/*
Copyright (C) 2008-2009  Anders Hedstrom

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
#include "ComplexType.h"
#include <Sockets/XmlNode.h>
#include "WsdlException.h"
#include "Element.h"
#include <Sockets/Debug.h>
#include "Split.h"
#include "Wsdl.h"
#include "Attribute.h"
#include "SimpleType.h"
#include "Application.h"
#include <iostream>


std::map<std::string, bool> ComplexType::m_b_generated;


ComplexType::ComplexType(const Wsdl& wsdl, const XmlNode& node, const std::string& name, const std::string& ns_prefix)
: TypeBase(wsdl, node, ns_prefix, Split(name).Type)
, m_name(!name.empty() ? name : node.PropertyExists("name") ? node.GetProperty("name") : "")
, m_b_is_sequence(false)
, m_b_is_all(false)
, m_b_complexContent(false)
, m_b_is_array(false)
, m_b_is_mixed(false)
{
	XmlNode start = Start();
	if (m_node.PropertyExists("mixed") && m_node.GetProperty("mixed") == "true")
	{
		m_b_is_mixed = true;
	}
	if (start.PropertyExists("base"))
	{
		m_extends = start.GetProperty("base");
		std::string ns_href = m_node.GetNsMap()[Split(m_extends).NS];
		if (ns_href.empty())
			m_extends_cls = Split(m_extends).CName1up;
		else
			m_extends_cls = m_wsdl.NamespaceFromHref(ns_href) + "::" + Split(m_extends).CName1up;
	}
	// check 'sequence'
	{
		XmlNode n(start, "sequence");
		if (n)
			m_b_is_sequence = true;
	}
	// check 'all'
	{
		XmlNode n(start, "all");
		if (n)
			m_b_is_all = true;
	}
	//
	{

		XmlNode n(m_node, "complexContent");
		if (n)
		{
			m_b_complexContent = true;
			XmlNode n2(n, "restriction");
			if (n2)
			{
				if (n2.PropertyExists("base") && Split(n2.GetProperty("base")).Type == "Array")
				{
					XmlNode n3(n2, "attribute");
					if (n3 && n3.PropertyExists("arrayType"))
					{
						std::string tmp = n3.GetProperty("arrayType");
						Parse pa(tmp, "[]");
						tmp = pa.getword();
						SetType( tmp );
						m_b_is_array = true;
					}
				}
			}
		}
	}
}


ComplexType::~ComplexType()
{
}


int ComplexType::NumberOfMembers() const
{
	XmlNode n(Start(), m_b_is_sequence ? "sequence" : "all");
	int q = 0;
	if (n)
	{
		XmlNode n2(n, "element");
		while (n2)
		{
			q++;
			++n2;
		}
	}
	return q;
}


bool ComplexType::GetRef(std::string& ref) const
{
	XmlNode n(Start(), m_b_is_sequence ? "sequence" : "all");
	if (n)
	{
		XmlNode n2(n, "element");
		while (n2)
		{
			if (n2.PropertyExists("ref"))
			{
				ref = n2.GetProperty("ref");
				return true;
			}
			++n2;
		}
	}
	return false;
}


std::list<std::pair<std::string, XmlNode> > ComplexType::GetMembers() const
{
	std::list<std::pair<std::string, XmlNode> > vec;
	XmlNode n1(Start(), m_b_is_sequence ? "sequence" : "all");
	if (n1)
	{
		XmlNode n(n1, "element");
		while (n)
		{
			if (n.PropertyExists("ref"))
			{
				std::string name = n.GetProperty("ref");
				vec.push_back(std::pair<std::string, XmlNode>(name, n));
			}
			else
			{
				std::string name = n.GetProperty("name");
				vec.push_back(std::pair<std::string, XmlNode>(name, n));
			}
			++n;
		}
	}
	return vec;
}


std::list<std::pair<std::string, XmlNode> > ComplexType::GetAttributes() const
{
	std::list<std::pair<std::string, XmlNode> > vec;
	XmlNode n(Start(), "attribute");
	while (n)
	{
		std::string name = n.GetProperty("name");
		vec.push_back(std::pair<std::string, XmlNode>(name, n));
		++n;
	}
	return vec;
}


const std::string ComplexType::Documentation() const
{
	return TypeBase::Documentation(m_node);
}


const std::string ComplexType::CName() const
{
	return Split(m_name).CName1up;
}


bool ComplexType::OkToGenerate() const
{
	Debug deb("OkToGenerate for '" + Name() + "' in ns '" + NsHref() + "'");
	if (m_b_complexContent)
	{
		if (m_b_is_array)
		{
			try
			{
				XmlNode n = m_wsdl.FindComplexType( m_ns_href, Type() );
				ComplexType t(m_wsdl, n, Split(Type()).Type, Split(Type()).NS);
				if (!t.IsGenerated())
				{
					return false;
				}
			}
			catch (const Exception& )
			{
			}
			return true;
		}
	}
	std::list<std::pair<std::string, XmlNode> > members = GetMembers();
	std::list<std::pair<std::string, XmlNode> > attrs = GetAttributes();
	for (std::list<std::pair<std::string, XmlNode> >::const_iterator it = members.begin(); it != members.end(); it++)
	{
		const std::pair<std::string, XmlNode>& ref = *it;
		Element el(m_wsdl, ref.second, m_b_is_sequence, false, "", NsHref());
		if (el.IsComplexType())
		{
			XmlNode n = m_wsdl.FindComplexType( el.NsHref(), el.Type() );
			ComplexType t(m_wsdl, n, Split(el.Type()).Type, Split(el.Type()).NS);
			std::string parent = m_wsdl.GetParent(NsHref(), t.Name());
			if (!t.IsGenerated() && parent != Name())
			{
				return false;
			}
		}
		else
		{
		}
	}
	return true;
}


void ComplexType::SetGenerated(bool x)
{
	m_b_generated[m_name] = x; 
}


bool ComplexType::IsGenerated() const
{
	std::map<std::string, bool>::const_iterator it = m_b_generated.find(m_name);

	if (it != m_b_generated.end())
	{
		return true;
	}
	return false;
}


void ComplexType::CreateInterface(FILE *fil, int lvl) const
{
	std::string prefix;
	for (int i = 0; i < lvl; i++)
		prefix += "\t";
	if (m_b_complexContent)
	{
		if (m_b_is_array)
		{
			fprintf(fil, "%stypedef %s<%s> %s;\n", prefix.c_str(), TypeBase::CType().c_str(), m_wsdl.GetApplication().ListType().c_str(), CName().c_str());
		}
	}

	std::list<std::pair<std::string, XmlNode> > members = GetMembers();
	std::list<std::pair<std::string, XmlNode> > attrs = GetAttributes();

	// always generate even if empty class
/*
	if (members.empty() && attrs.empty() && !m_b_is_mixed)
		return;
*/

	// parent
	std::string parent = m_wsdl.GetParent(NsHref(), Name());

	// class
	if (m_extends.empty())
	{
		fprintf(fil, "%sclass %s\n%s{\n", prefix.c_str(), CName().c_str(), prefix.c_str());
	}
	else
	{
		fprintf(fil, "%sclass %s : public %s\n%s{\n", prefix.c_str(), CName().c_str(), m_extends_cls.c_str(), prefix.c_str());
	}

	// internal types
	try
	{
		const std::list<std::string>& ref = m_wsdl.ComplexTypes(NsHref());
		for (std::list<std::string>::const_iterator it = ref.begin(); it != ref.end(); it++)
		{
			std::string parent = m_wsdl.GetParent(NsHref(), *it);
			if (parent == Name())
			{
				fprintf(fil, "%spublic:\n", prefix.c_str());
				XmlNode n = m_wsdl.FindComplexType( m_ns_href, *it );
				ComplexType t(m_wsdl, n, *it, NsHref());
				t.CreateInterface(fil, lvl + 1);
			}
		}
	} catch (const Exception&) {}
	try
	{
		const std::list<std::string>& ref = m_wsdl.SimpleTypes(NsHref());
		for (std::list<std::string>::const_iterator it = ref.begin(); it != ref.end(); it++)
		{
			std::string parent = m_wsdl.GetParent(NsHref(), *it);
			if (parent == Name())
			{
				fprintf(fil, "%spublic:\n", prefix.c_str());
				XmlNode n = m_wsdl.FindSimpleType( m_ns_href, *it );
				SimpleType t(m_wsdl, n, *it, NsHref());
				t.CreateInterface(fil, lvl + 1);
			}
		}
	} catch (const Exception&) {}

	// begin methods
	fprintf(fil, "%spublic:\n", prefix.c_str());
	// default constructor
	fprintf(fil, "%s\t%s();\n", prefix.c_str(), CName().c_str());
	// create from XmlNode constructor
	fprintf(fil, "%s\t%s(const XmlNode& it);\n", prefix.c_str(), CName().c_str());
	// destructor
	fprintf(fil, "%s\tvirtual ~%s() {}\n", prefix.c_str(), CName().c_str());
	fprintf(fil, "\n");
	// clear
	fprintf(fil, "%s\tvoid clear();\n", prefix.c_str());
	fprintf(fil, "\n");
	// From Xml
	fprintf(fil, "%s\tvoid FromXml(const XmlNode& it);\n", prefix.c_str());
	fprintf(fil, "\n");
	// ToString
	fprintf(fil, "%s\tconst std::string ToString(const std::string& name = \"\") const;\n", prefix.c_str());
	fprintf(fil, "\n");
	// IsSet
	{
		fprintf(fil, "%s\tbool IsSet() const;\n", prefix.c_str());
		fprintf(fil, "\n");
	}
	// get/set methods
	for (std::list<std::pair<std::string, XmlNode> >::const_iterator it = members.begin(); it != members.end(); it++)
	{
		const std::pair<std::string, XmlNode>& ref = *it;
		Element el(m_wsdl, ref.second, m_b_is_sequence, false, "", NsHref());
		if (el.IsVector() || el.IsComplexType())
		{
			fprintf(fil, "%s\tconst %s Get%s() const;\n", prefix.c_str(), el.CType_out().c_str(), el.CName1up().c_str());
			fprintf(fil, "%s\t%s Get%s();\n", prefix.c_str(), el.CType_out().c_str(), el.CName1up().c_str());
			fprintf(fil, "%s\tvoid Set%s(%s %s);\n",
				prefix.c_str(), el.CName1up().c_str(), el.CType_in().c_str(), el.CName().c_str());
		}
		else
		{
			fprintf(fil, "%s\t%s Get%s() const;\n", prefix.c_str(), el.CType_out().c_str(), el.CName1up().c_str());
			fprintf(fil, "%s\tvoid Set%s(%s %s);\n", prefix.c_str(), el.CName1up().c_str(),
				el.CType_in().c_str(), el.CName().c_str());
		}
		fprintf(fil, "\n");
	}
	for (std::list<std::pair<std::string, XmlNode> >::const_iterator it = attrs.begin(); it != attrs.end(); it++)
	{
		const std::pair<std::string, XmlNode>& ref = *it;
		Attribute attr(m_wsdl, ref.second, ref.first, NsHref());

		fprintf(fil, "%s\t%s Get%s() const;\n", prefix.c_str(), attr.CType_out().c_str(), attr.CName1up().c_str());
//		if (!attr.FixedIsSet())
		{
			fprintf(fil, "%s\tvoid Set%s(%s %s);\n", prefix.c_str(), attr.CName1up().c_str(),
				attr.CType_in().c_str(), attr.CName().c_str());
		}
	}
	if (!attrs.empty())
		fprintf(fil, "\n");
	if (m_b_is_mixed)
	{
		fprintf(fil, "%s\tconst std::string& Get() const;\n", prefix.c_str());
		fprintf(fil, "%s\tvoid Set(const std::string&);\n", prefix.c_str());
		fprintf(fil, "\n");
	}
	// is_set methods
	for (std::list<std::pair<std::string, XmlNode> >::const_iterator it = members.begin(); it != members.end(); it++)
	{
		const std::pair<std::string, XmlNode>& ref = *it;
		Element el(m_wsdl, ref.second, m_b_is_sequence, false, "", NsHref());
		if (el.IsVector() || el.IsComplexType())
		{
			// _is_set is calculated
			fprintf(fil, "%s\tbool %sIsSet() const;\n", prefix.c_str(), el.CName1up().c_str());
		}
		else
		{
			fprintf(fil, "%s\tbool %sIsSet() const { return _element_%s_is_set; }\n", prefix.c_str(), el.CName1up().c_str(), el.CName().c_str());
		}
	}
	if (!members.empty())
		fprintf(fil, "\n");
	for (std::list<std::pair<std::string, XmlNode> >::const_iterator it = attrs.begin(); it != attrs.end(); it++)
	{
		const std::pair<std::string, XmlNode>& ref = *it;
		Attribute attr(m_wsdl, ref.second, ref.first, NsHref());

		fprintf(fil, "%s\tbool %sIsSet() const { return _attribute_%s_is_set; }\n", prefix.c_str(), attr.CName1up().c_str(), attr.CName().c_str());
	}
	if (!attrs.empty())
		fprintf(fil, "\n");
	// private
	fprintf(fil, "%sprivate:\n", prefix.c_str());
	// elements
	for (std::list<std::pair<std::string, XmlNode> >::const_iterator it = members.begin(); it != members.end(); it++)
	{
		const std::pair<std::string, XmlNode>& ref = *it;
		Element el(m_wsdl, ref.second, m_b_is_sequence, false, "", NsHref());
		std::string doc = el.Documentation();
		if (!doc.empty())
			fprintf(fil, "%s\t/** %s */\n", prefix.c_str(), doc.c_str());
		fprintf(fil, "%s\t%s _element_%s;\n", prefix.c_str(), el.CType_decl().c_str(), el.CName().c_str());
	}
	for (std::list<std::pair<std::string, XmlNode> >::const_iterator it = attrs.begin(); it != attrs.end(); it++)
	{
		const std::pair<std::string, XmlNode>& ref = *it;
		Attribute attr(m_wsdl, ref.second, ref.first, NsHref());

		fprintf(fil, "%s\t%s _attribute_%s;\n", prefix.c_str(), attr.CType_decl().c_str(), attr.CName().c_str());
	}
	if (m_b_is_mixed)
	{
		fprintf(fil, "%s\tstd::string m_mixed_content;\n", prefix.c_str());
	}
	// is_set members
	for (std::list<std::pair<std::string, XmlNode> >::const_iterator it = members.begin(); it != members.end(); it++)
	{
		const std::pair<std::string, XmlNode>& ref = *it;
		Element el(m_wsdl, ref.second, m_b_is_sequence, false, "", NsHref());
		if (el.IsVector() || el.IsComplexType())
		{
			// _is_set is calculated
		}
		else
		{
			fprintf(fil, "%s\tbool _element_%s_is_set;\n", prefix.c_str(), el.CName().c_str());
		}
	}
	for (std::list<std::pair<std::string, XmlNode> >::const_iterator it = attrs.begin(); it != attrs.end(); it++)
	{
		const std::pair<std::string, XmlNode>& ref = *it;
		Attribute attr(m_wsdl, ref.second, ref.first, NsHref());

		fprintf(fil, "%s\tbool _attribute_%s_is_set;\n", prefix.c_str(), attr.CName().c_str());
	}
	//
	fprintf(fil, "%s};\n", prefix.c_str()); // End of class '%s'\n", CName().c_str());
	fprintf(fil, "\n");
}


void ComplexType::reset_members(std::list<std::pair<std::string, XmlNode> >& members, std::list<std::pair<std::string, XmlNode> >& attrs, FILE *fil) const
{
	std::string prefix = ": ";
	if (!m_extends.empty())
	{
		fprintf(fil, "%s%s()\n", prefix.c_str(), m_extends_cls.c_str());
		prefix = ", ";
	}
	for (std::list<std::pair<std::string, XmlNode> >::const_iterator it = members.begin(); it != members.end(); it++)
	{
		const std::pair<std::string, XmlNode>& ref = *it;
		Element el(m_wsdl, ref.second, m_b_is_sequence, false, "", NsHref());
		std::string def = el.DefaultValue();
		if (el.IsBuiltin() && el.HasDefault())
		{
			if (el.TypeClass() == T_NUMBER)
				fprintf(fil, "%s_element_%s(%s)\n", prefix.c_str(), el.CName().c_str(), el.Default().c_str());
			else
				fprintf(fil, "%s_element_%s(\"%s\")\n", prefix.c_str(), el.CName().c_str(), el.Default().c_str());
			prefix = ", ";
		}
		else
		if (el.IsBuiltin() && !def.empty())
		{
			fprintf(fil, "%s_element_%s(%s)\n", prefix.c_str(), el.CName().c_str(), el.DefaultValue().c_str());
			prefix = ", ";
		}
	}
	for (std::list<std::pair<std::string, XmlNode> >::const_iterator it = members.begin(); it != members.end(); it++)
	{
		const std::pair<std::string, XmlNode>& ref = *it;
		Element el(m_wsdl, ref.second, m_b_is_sequence, false, "", NsHref());
		if (el.IsVector() || el.IsComplexType())
		{
			// _is_set is calculated
		}
		else
		if (el.HasDefault())
		{
			fprintf(fil, "%s_element_%s_is_set(true)\n", prefix.c_str(), el.CName().c_str());
			prefix = ", ";
		}
		else
		{
			fprintf(fil, "%s_element_%s_is_set(false)\n", prefix.c_str(), el.CName().c_str());
			prefix = ", ";
		}
	}
	for (std::list<std::pair<std::string, XmlNode> >::const_iterator it = attrs.begin(); it != attrs.end(); it++)
	{
		const std::pair<std::string, XmlNode>& ref = *it;
		Attribute attr(m_wsdl, ref.second, "", NsHref());
		if (attr.DefaultIsSet())
		{
			if (attr.TypeClass() == T_NUMBER || attr.TypeClass() == T_BOOLEAN)
				fprintf(fil, "%s_attribute_%s(%s)\n", prefix.c_str(), attr.CName().c_str(), attr.Default().c_str());
			else
				fprintf(fil, "%s_attribute_%s(\"%s\")\n", prefix.c_str(), attr.CName().c_str(), attr.Default().c_str());
			prefix = ", ";
		}
		else
		if (attr.FixedIsSet())
		{
			if (attr.TypeClass() == T_NUMBER || attr.TypeClass() == T_BOOLEAN)
				fprintf(fil, "%s_attribute_%s(%s)\n", prefix.c_str(), attr.CName().c_str(), attr.Fixed().c_str());
			else
				fprintf(fil, "%s_attribute_%s(\"%s\")\n", prefix.c_str(), attr.CName().c_str(), attr.Fixed().c_str());
			prefix = ", ";
		}
		else
		{
			std::string def = attr.DefaultValue();
			if (attr.IsBuiltin() && !def.empty())
			{
				fprintf(fil, "%s_attribute_%s(%s)\n", prefix.c_str(), attr.CName().c_str(), attr.DefaultValue().c_str());
				prefix = ", ";
			}
		}
	}
	for (std::list<std::pair<std::string, XmlNode> >::const_iterator it = attrs.begin(); it != attrs.end(); it++)
	{
		const std::pair<std::string, XmlNode>& ref = *it;
		Attribute attr(m_wsdl, ref.second, "", NsHref());
		if (attr.DefaultIsSet() || attr.FixedIsSet())
			fprintf(fil, "%s_attribute_%s_is_set(true)\n", prefix.c_str(), attr.CName().c_str());
		else
			fprintf(fil, "%s_attribute_%s_is_set(false)\n", prefix.c_str(), attr.CName().c_str());
		prefix = ", ";
	}
}


void ComplexType::clear_members(std::list<std::pair<std::string, XmlNode> >& members, std::list<std::pair<std::string, XmlNode> >& attrs, FILE *fil) const
{
	if (!m_extends.empty())
	{
		fprintf(fil, "\t%s::clear();\n", m_extends_cls.c_str());
	}
	for (std::list<std::pair<std::string, XmlNode> >::const_iterator it = members.begin(); it != members.end(); it++)
	{
		const std::pair<std::string, XmlNode>& ref = *it;
		Element el(m_wsdl, ref.second, m_b_is_sequence, false, "", NsHref());
		std::string def = el.DefaultValue();
		if (el.IsVector())
		{
			// %! clear vector
			fprintf(fil, "\twhile (!_element_%s.empty())\n", el.CName().c_str());
			fprintf(fil, "\t\t_element_%s.erase( _element_%s.begin() );\n", el.CName().c_str(), el.CName().c_str());
		}
		else
		if (el.IsSimpleType())
		{
			fprintf(fil, "\t// simpleType:\n");
			fprintf(fil, "\t// _element_%s.clear();\n", el.CName().c_str());
		}
		else
		if (el.IsComplexType())
		{
			fprintf(fil, "\t_element_%s.clear();\n", el.CName().c_str());
		}
		else
		if (el.IsBuiltin() && el.HasDefault())
		{
			if (el.TypeClass() == T_NUMBER)
				fprintf(fil, "\t_element_%s = %s;\n", el.CName().c_str(), el.Default().c_str());
			else
				fprintf(fil, "\t_element_%s = \"%s\";\n", el.CName().c_str(), el.Default().c_str());
		}
		else
		if (el.IsBuiltin() && !def.empty())
		{
			fprintf(fil, "\t_element_%s = %s;\n", el.CName().c_str(), el.DefaultValue().c_str());
		}
		else
		{
			if (el.TypeClass() == T_NUMBER)
				fprintf(fil, "\t_element_%s = 0;\n", el.CName().c_str());
			else
			if (el.TypeClass() == T_BOOLEAN)
				fprintf(fil, "\t_element_%s = false;\n", el.CName().c_str());
			else
				fprintf(fil, "\t_element_%s.clear();\n", el.CName().c_str());
		}
	}
	for (std::list<std::pair<std::string, XmlNode> >::const_iterator it = members.begin(); it != members.end(); it++)
	{
		const std::pair<std::string, XmlNode>& ref = *it;
		Element el(m_wsdl, ref.second, m_b_is_sequence, false, "", NsHref());
		if (el.IsVector() || el.IsComplexType())
		{
			// _is_set is calculated
		}
		else
		if (el.HasDefault())
		{
			fprintf(fil, "\t_element_%s_is_set = true;\n", el.CName().c_str());
		}
		else
		{
			fprintf(fil, "\t_element_%s_is_set = false;\n", el.CName().c_str());
		}
	}
	for (std::list<std::pair<std::string, XmlNode> >::const_iterator it = attrs.begin(); it != attrs.end(); it++)
	{
		const std::pair<std::string, XmlNode>& ref = *it;
		Attribute attr(m_wsdl, ref.second, "", NsHref());
		if (attr.DefaultIsSet())
		{
			if (attr.TypeClass() == T_NUMBER || attr.TypeClass() == T_BOOLEAN)
				fprintf(fil, "\t_attribute_%s = %s;\n", attr.CName().c_str(), attr.Default().c_str());
			else
				fprintf(fil, "\t_attribute_%s = \"%s\";\n", attr.CName().c_str(), attr.Default().c_str());
		}
		else
		if (attr.FixedIsSet())
		{
			if (attr.TypeClass() == T_NUMBER || attr.TypeClass() == T_BOOLEAN)
				fprintf(fil, "\t_attribute_%s = %s;\n", attr.CName().c_str(), attr.Fixed().c_str());
			else
				fprintf(fil, "\t_attribute_%s = \"%s\";\n", attr.CName().c_str(), attr.Fixed().c_str());
		}
		else
		{
			std::string def = attr.DefaultValue();
			if (attr.IsBuiltin() && !def.empty())
			{
				fprintf(fil, "\t_attribute_%s = %s;\n", attr.CName().c_str(), attr.DefaultValue().c_str());
			}
			else
			{
				if (attr.TypeClass() == T_NUMBER)
					fprintf(fil, "\t_attribute_%s = 0;\n", attr.CName().c_str());
				else
				if (attr.TypeClass() == T_BOOLEAN)
					fprintf(fil, "\t_attribute_%s = false;\n", attr.CName().c_str());
				else
					fprintf(fil, "\t_attribute_%s.clear();\n", attr.CName().c_str());
			}
		}
	}
	for (std::list<std::pair<std::string, XmlNode> >::const_iterator it = attrs.begin(); it != attrs.end(); it++)
	{
		const std::pair<std::string, XmlNode>& ref = *it;
		Attribute attr(m_wsdl, ref.second, "", NsHref());
		if (attr.DefaultIsSet() || attr.FixedIsSet())
			fprintf(fil, "\t_attribute_%s_is_set = true;\n", attr.CName().c_str());
		else
			fprintf(fil, "\t_attribute_%s_is_set = false;\n", attr.CName().c_str());
	}
}


void ComplexType::from_xml(std::list<std::pair<std::string, XmlNode> >& members, std::list<std::pair<std::string, XmlNode> >& attrs, FILE *fil) const
{	
	if (!m_extends.empty())
	{
		fprintf(fil, "\t%s::FromXml(it);\n", m_extends_cls.c_str());
	}
	for (std::list<std::pair<std::string, XmlNode> >::const_iterator it = members.begin(); it != members.end(); it++)
	{
		const std::pair<std::string, XmlNode>& ref = *it;
		Element el(m_wsdl, ref.second, m_b_is_sequence, false, "", NsHref());
		if (!el.IsOptional())
		{
			fprintf(fil, "\tif (!it.Exists(\"%s\"))\n", el.Name().c_str());
			fprintf(fil, "\t\tthrow XmlException(\"missing required node in element '%s': %s\");\n", m_name.c_str(), el.Name().c_str());
		}
		if (el.IsVector())
		{
			fprintf(fil, "\t{\n");
			fprintf(fil, "\t\tXmlNode n(it, \"%s\");\n", el.Name().c_str());
			fprintf(fil, "\t\twhile (n)\n");
			fprintf(fil, "\t\t{\n");
			if (el.IsComplexType())
			{
				fprintf(fil, "\t\t\t%s tmp(n);\n", el.CType().c_str());
				fprintf(fil, "\t\t\t_element_%s.push_back( tmp );\n", el.CName().c_str());
			}
			else
			{
//				fprintf(fil, "\t\t\t// el.BaseType() == '%s'\n", el.BaseType().c_str());
				if (el.BaseType() == "std::string")
				{
					fprintf(fil, "\t\t\t_element_%s.push_back( n.GetContent() );\n", el.CName().c_str());
				}
				else
				if (el.BaseType() == "double")
				{
					fprintf(fil, "\t\t\t_element_%s.push_back( atof(n.GetContent().c_str()) );\n", el.CName().c_str());
				}
				else
				if (el.BaseType() == "bool")
				{
					fprintf(fil, "\t\t\t_element_%s.push_back( n.GetContent() == \"true\" || atoi(n.GetContent().c_str()) != 0 || n.GetContent() == \"yes\" );\n", el.CName().c_str());
				}
				else
				{
					fprintf(fil, "\t\t\t_element_%s.push_back( atoi(n.GetContent().c_str()) );\n", el.CName().c_str());
				}
			}
			fprintf(fil, "\t\t\t++n;\n");
			fprintf(fil, "\t\t}\n");
			fprintf(fil, "\t}\n");
		}
		else
		if (el.IsComplexType())
		{
			fprintf(fil, "\t{\n");
			fprintf(fil, "\t\tXmlNode n(it, \"%s\");\n", el.Name().c_str());
			fprintf(fil, "\t\tif (n)\n");
			fprintf(fil, "\t\t\t_element_%s.FromXml( n );\n", el.CName().c_str());
			fprintf(fil, "\t}\n");
		}
		else
		{
//			fprintf(fil, "\t// el.BaseType() == '%s'\n", el.BaseType().c_str());
			if (el.BaseType() == "std::string")
			{
				if (el.IsOptional())
					fprintf(fil, "\tif (it.Exists(\"%s\"))\n\t", el.Name().c_str());
				fprintf(fil, "\tSet%s( it[\"%s\"] );\n", el.CName1up().c_str(), el.Name().c_str());
			}
			else
			if (el.BaseType() == "double")
			{
				if (el.IsOptional())
					fprintf(fil, "\tif (it.Exists(\"%s\"))\n\t", el.Name().c_str());
				fprintf(fil, "\tSet%s( atof(std::string(it[\"%s\"]).c_str()) );\n", el.CName1up().c_str(), el.Name().c_str());
			}
			else
			if (el.BaseType() == "bool")
			{
				if (el.IsOptional())
					fprintf(fil, "\tif (it.Exists(\"%s\"))\n", el.Name().c_str());
				fprintf(fil, "\t{\n");
				fprintf(fil, "\t\tstd::string tmp = std::string(it[\"%s\"]);\n", el.Name().c_str());
				fprintf(fil, "\t\tSet%s( tmp == \"true\" || atoi(tmp.c_str()) != 0 || tmp == \"yes\" );\n", el.CName1up().c_str());
				fprintf(fil, "\t}\n");
			}
			else
			{
				if (el.IsOptional())
					fprintf(fil, "\tif (it.Exists(\"%s\"))\n\t", el.Name().c_str());
				fprintf(fil, "\tSet%s( atoi(std::string(it[\"%s\"]).c_str()) );\n", el.CName1up().c_str(), el.Name().c_str());
			}
		}
	}
	for (std::list<std::pair<std::string, XmlNode> >::const_iterator it = attrs.begin(); it != attrs.end(); it++)
	{
		const std::pair<std::string, XmlNode>& ref = *it;
		Attribute attr(m_wsdl, ref.second, "", NsHref());

//		fprintf(fil, "\t// attr.BaseType() == '%s'\n", attr.BaseType().c_str());
		if (attr.BaseType() == "std::string")
		{
			if (attr.IsOptional())
			{
				fprintf(fil, "\tif (it.PropertyExists(\"%s\"))\n\t", attr.Name().c_str());
			}
			fprintf(fil, "\tSet%s(it.GetProperty(\"%s\"));\n", attr.CName1up().c_str(), attr.Name().c_str());
		}
		else
		if (attr.BaseType() == "double")
		{
			if (attr.IsOptional())
			{
				fprintf(fil, "\tif (it.PropertyExists(\"%s\"))\n\t", attr.Name().c_str());
			}
			fprintf(fil, "\tSet%s( atof(it.GetProperty(\"%s\").c_str()) );\n", attr.CName1up().c_str(), attr.Name().c_str());
		}
		else
		if (attr.BaseType() == "bool")
		{
			if (attr.IsOptional())
			{
				fprintf(fil, "\tif (it.PropertyExists(\"%s\"))\n", attr.Name().c_str());
			}
			fprintf(fil, "\t{\n");
			fprintf(fil, "\t\tstd::string tmp = it.GetProperty(\"%s\");\n", attr.Name().c_str());
			fprintf(fil, "\t\tSet%s( tmp == \"true\" || atoi(tmp.c_str()) != 0 || tmp == \"yes\" );\n", attr.CName1up().c_str());
			fprintf(fil, "\t}\n");
		}
		else
		{
			if (attr.IsOptional())
			{
				fprintf(fil, "\tif (it.PropertyExists(\"%s\"))\n\t", attr.Name().c_str());
			}
			fprintf(fil, "\tSet%s( atoi(it.GetProperty(\"%s\").c_str()) );\n", attr.CName1up().c_str(), attr.Name().c_str());
		}
	}
	if (m_b_is_mixed)
	{
		fprintf(fil, "\tm_mixed_content = it.GetContent();\n");
	}
}


void ComplexType::to_string(std::list<std::pair<std::string, XmlNode> >& members, std::list<std::pair<std::string, XmlNode> >& attrs, FILE *fil) const
{
	fprintf(fil, "\tstd::string r;\n");
	if (attrs.empty())
	{
		fprintf(fil, "\tif (!name.empty())\n\t");
		fprintf(fil, "\tr += \"<\" + name + \">\";\n");
	}
	else
	{
		fprintf(fil, "\tif (!name.empty())\n\t");
		fprintf(fil, "\tr += \"<\" + name;\n");

		for (std::list<std::pair<std::string, XmlNode> >::const_iterator it = attrs.begin(); it != attrs.end(); it++)
		{
			const std::pair<std::string, XmlNode>& ref = *it;
			Attribute attr(m_wsdl, ref.second, "", NsHref());

			if (attr.IsOptional())
			{
				fprintf(fil, "\tif (%sIsSet())\n\t", attr.CName1up().c_str());
			}
			if (attr.BaseType() == "std::string")
				fprintf(fil, "\tr += \" %s='\" + Get%s() + \"'\";\n", attr.Name().c_str(), attr.CName1up().c_str());
			else
			{
				fprintf(fil, "{\n");
				// %!
				fprintf(fil, "\t\tr = r + \" %s='\" + Utility::l2string(Get%s()) + \"'\";\n", attr.Name().c_str(), attr.CName1up().c_str());
				fprintf(fil, "\t}\n");
			}
		}
		fprintf(fil, "\tr += \">\";\n");
	}
	if (!m_extends.empty())
	{
		fprintf(fil, "\tr += %s::ToString();\n", m_extends_cls.c_str());
	}
	for (std::list<std::pair<std::string, XmlNode> >::const_iterator it = members.begin(); it != members.end(); it++)
	{
		const std::pair<std::string, XmlNode>& ref = *it;
		Element el(m_wsdl, ref.second, m_b_is_sequence, false, "", NsHref());
		if (el.IsVector())
		{
			if (el.IsSimpleType())
			{
				fprintf(fil, "\tfor (%s<%s>::const_iterator it = _element_%s.begin(); it != _element_%s.end(); it++)\n",
					m_wsdl.GetApplication().ListType().c_str(),
					el.BaseType().c_str(), el.CName().c_str(), el.CName().c_str());
				fprintf(fil, "\t{\n");

				if (el.BaseType() == "std::string")
					fprintf(fil, "\t\tr += \"<%s>\" + *it + \"</%s>\";\n",
						el.Name().c_str(), el.Name().c_str());
				else
					// %!
					fprintf(fil, "\t\t\tr = r + \"<%s>\" + Utility::l2string(*it) + \"</%s>\";\n",
						el.Name().c_str(), el.Name().c_str());
			}
			else
			if (el.IsComplexType())
			{
				fprintf(fil, "\tfor (%s<%s>::const_iterator it = _element_%s.begin(); it != _element_%s.end(); it++)\n",
					m_wsdl.GetApplication().ListType().c_str(),
					el.CType().c_str(), el.CName().c_str(), el.CName().c_str());
				fprintf(fil, "\t{\n");

				fprintf(fil, "\t\tconst %s& ref = *it;\n", el.CType().c_str());
				fprintf(fil, "\t\tr += ref.ToString(\"%s\");\n", el.Name().c_str());
			}
			else
			if (el.BaseType() == "std::string")
			{
				fprintf(fil, "\tfor (%s<%s>::const_iterator it = _element_%s.begin(); it != _element_%s.end(); it++)\n",
					m_wsdl.GetApplication().ListType().c_str(),
					el.BaseType().c_str(), el.CName().c_str(), el.CName().c_str());
				fprintf(fil, "\t{\n");

				fprintf(fil, "\t\tr += \"<%s>\" + *it + \"</%s>\";\n",
					el.Name().c_str(), el.Name().c_str());
			}
			else
			{
				fprintf(fil, "\tfor (%s<%s>::const_iterator it = _element_%s.begin(); it != _element_%s.end(); it++)\n",
					m_wsdl.GetApplication().ListType().c_str(),
					el.CType().c_str(), el.CName().c_str(), el.CName().c_str());
				fprintf(fil, "\t{\n");

				fprintf(fil, "\t\t{\n");
				// %!
				fprintf(fil, "\t\t\tr = r + \"<%s>\" + Utility::l2string(*it) + \"</%s>\";\n",
					el.Name().c_str(), el.Name().c_str());
				fprintf(fil, "\t\t}\n");
			}

			fprintf(fil, "\t}\n");
		}
		else
		if (el.IsComplexType())
		{
			fprintf(fil, "\tr += _element_%s.ToString(\"%s\");\n", el.CName().c_str(), el.Name().c_str());
		}
		else
		if (el.BaseType() == "std::string")
		{
			if (el.IsOptional())
				fprintf(fil, "\tif (_element_%s_is_set)\n\t", el.CName().c_str());
			fprintf(fil, "\tr += \"<%s>\" + _element_%s + \"</%s>\";\n",
				el.Name().c_str(), el.CName().c_str(), el.Name().c_str());
		}
		else
		{
			if (el.IsOptional())
				fprintf(fil, "\tif (_element_%s_is_set)\n", el.CName().c_str());
			fprintf(fil, "\t{\n");
			// %!
			fprintf(fil, "\t\tr = r + \"<%s>\" + Utility::l2string(_element_%s) + \"</%s>\";\n",
				el.Name().c_str(), el.CName().c_str(), el.Name().c_str());
			fprintf(fil, "\t}\n");
		}
	}
	if (m_b_is_mixed)
	{
		fprintf(fil, "\tr += m_mixed_content;\n");
	}
	fprintf(fil, "\tif (!name.empty())\n\t");
	fprintf(fil, "\tr += \"</\" + name + \">\";\n");
	fprintf(fil, "\treturn r;\n");
}


void ComplexType::CreateImplementation(FILE *fil) const
{
	std::list<std::pair<std::string, XmlNode> > members = GetMembers();
	std::list<std::pair<std::string, XmlNode> > attrs = GetAttributes();
/*
	if (members.empty() && attrs.empty() && !m_b_is_mixed)
		return;
*/

	std::string parent = m_wsdl.GetParent(NsHref(), Name());
	std::string tmp;
	while (!parent.empty())
	{
		tmp = Split(parent).CName1up + "::" + tmp;
		parent = m_wsdl.GetParent(NsHref(), parent);
	}
	std::string classname = tmp + CName();

	// default constructor
	fprintf(fil, "// -----------------------------------------------------------------------------\n");
	fprintf(fil, "%s::%s()\n", classname.c_str(), CName().c_str());
	reset_members(members, attrs, fil);
	fprintf(fil, "{\n");
	fprintf(fil, "}\n\n\n");
	// XmlNode constructor
	fprintf(fil, "// -----------------------------------------------------------------------------\n");
	fprintf(fil, "%s::%s(const XmlNode& it)\n", classname.c_str(), CName().c_str());
	reset_members(members, attrs, fil);
	fprintf(fil, "{\n");
	fprintf(fil, "\tFromXml(it);\n");
	fprintf(fil, "}\n\n\n");
	// clear
	fprintf(fil, "// -----------------------------------------------------------------------------\n");
	fprintf(fil, "void %s::clear()\n{\n", classname.c_str());
	clear_members(members, attrs, fil);
	fprintf(fil, "}\n\n\n");
	// From Xml
	fprintf(fil, "// -----------------------------------------------------------------------------\n");
	fprintf(fil, "void %s::FromXml(const XmlNode& it)\n{\n", classname.c_str());
	from_xml(members, attrs, fil);
	fprintf(fil, "}\n\n\n");
	// ToString
	fprintf(fil, "// -----------------------------------------------------------------------------\n");
	fprintf(fil, "const std::string %s::ToString(const std::string& name) const\n{\n", classname.c_str());
	to_string(members, attrs, fil);
	fprintf(fil, "}\n\n\n");
	// IsSet
	{
		fprintf(fil, "// -----------------------------------------------------------------------------\n");
		fprintf(fil, "bool %s::IsSet() const\n{\n", classname.c_str());
		std::string prefix = "\treturn ";
		for (std::list<std::pair<std::string, XmlNode> >::const_iterator it = members.begin(); it != members.end(); it++)
		{
			const std::pair<std::string, XmlNode>& ref = *it;
			Element el(m_wsdl, ref.second, m_b_is_sequence, false, "", NsHref());
			fprintf(fil, "%s%sIsSet()", prefix.c_str(), el.CName1up().c_str());
			prefix = " || ";
		}
		for (std::list<std::pair<std::string, XmlNode> >::const_iterator it = attrs.begin(); it != attrs.end(); it++)
		{
			const std::pair<std::string, XmlNode>& ref = *it;
			Attribute attr(m_wsdl, ref.second, "", NsHref());
			fprintf(fil, "%s%sIsSet()", prefix.c_str(), attr.CName1up().c_str());
			prefix = " || ";
		}
		if (m_b_is_mixed)
		{
			fprintf(fil, "%s!m_mixed_content.empty()", prefix.c_str());
			prefix = " || ";
		}
		fprintf(fil, ";\n");
		fprintf(fil, "}\n\n\n");
	}
	// get/set
	for (std::list<std::pair<std::string, XmlNode> >::const_iterator it = members.begin(); it != members.end(); it++)
	{
		const std::pair<std::string, XmlNode>& ref = *it;
		Element el(m_wsdl, ref.second, m_b_is_sequence, false, "", NsHref());
		fprintf(fil, "// -----------------------------------------------------------------------------\n");
		if (el.IsVector() || el.IsComplexType())
		{
			fprintf(fil, "const %s %s::Get%s() const\n{\n", el.CType_out().c_str(), classname.c_str(), el.CName1up().c_str());
			fprintf(fil, "\treturn _element_%s;\n", el.CName().c_str());
			fprintf(fil, "}\n\n\n");
			fprintf(fil, "%s %s::Get%s()\n{\n", el.CType_out().c_str(), classname.c_str(), el.CName1up().c_str());
			fprintf(fil, "\treturn _element_%s;\n", el.CName().c_str());
			fprintf(fil, "}\n\n\n");
			fprintf(fil, "void %s::Set%s(%s %s)\n{\n\t_element_%s = %s;\n}\n\n\n",
				classname.c_str(),
				el.CName1up().c_str(), el.CType_in().c_str(), el.CName().c_str(),
				el.CName().c_str(), el.CName().c_str());
		}
		else
		{
			fprintf(fil, "%s %s::Get%s() const\n{\n", el.CType_out().c_str(), classname.c_str(), el.CName1up().c_str());
			fprintf(fil, "\treturn _element_%s;\n", el.CName().c_str());
			fprintf(fil, "}\n\n\n");
			fprintf(fil, "void %s::Set%s(%s %s)\n{\n", classname.c_str(), el.CName1up().c_str(),
				el.CType_in().c_str(), el.CName().c_str());
			fprintf(fil, "\t_element_%s = %s;\n", el.CName().c_str(), el.CName().c_str());
			fprintf(fil, "\t_element_%s_is_set = true;\n", el.CName().c_str());
			fprintf(fil, "}\n\n\n");
		}
	}
	// is_set
	for (std::list<std::pair<std::string, XmlNode> >::const_iterator it = members.begin(); it != members.end(); it++)
	{
		const std::pair<std::string, XmlNode>& ref = *it;
		Element el(m_wsdl, ref.second, m_b_is_sequence, false, "", NsHref());
		if (el.IsVector() || el.IsComplexType())
		{
			// _is_set is calculated
			fprintf(fil, "// -----------------------------------------------------------------------------\n");
			fprintf(fil, "bool %s::%sIsSet() const\n{\n", classname.c_str(), el.CName1up().c_str());
			if (el.IsVector())
				fprintf(fil, "\treturn !_element_%s.empty();\n", el.CName().c_str());
			else
				fprintf(fil, "\treturn _element_%s.IsSet();\n", el.CName().c_str());
			fprintf(fil, "}\n\n\n");
		}
		else
		{
		}
	}
	// attribute get/set
	for (std::list<std::pair<std::string, XmlNode> >::const_iterator it = attrs.begin(); it != attrs.end(); it++)
	{
		const std::pair<std::string, XmlNode>& ref = *it;
		Attribute attr(m_wsdl, ref.second, "", NsHref());
		fprintf(fil, "// -----------------------------------------------------------------------------\n");
		{
			fprintf(fil, "%s %s::Get%s() const\n{\n", attr.CType_out().c_str(), classname.c_str(), attr.CName1up().c_str());
			fprintf(fil, "\treturn _attribute_%s;\n", attr.CName().c_str());
			fprintf(fil, "}\n\n\n");
//			if (!attr.FixedIsSet())
			{
				fprintf(fil, "void %s::Set%s(%s %s)\n{\n", classname.c_str(), attr.CName1up().c_str(),
					attr.CType_in().c_str(), attr.CName().c_str());
				fprintf(fil, "\t_attribute_%s = %s;\n", attr.CName().c_str(), attr.CName().c_str());
				fprintf(fil, "\t_attribute_%s_is_set = true;\n", attr.CName().c_str());
				fprintf(fil, "}\n\n\n");
			}
		}
	}
	if (m_b_is_mixed)
	{
		fprintf(fil, "// -----------------------------------------------------------------------------\n");
		fprintf(fil, "const std::string& %s::Get() const\n", classname.c_str());
		fprintf(fil, "{\n");
		fprintf(fil, "\treturn m_mixed_content;\n");
		fprintf(fil, "}\n\n\n");
		fprintf(fil, "// -----------------------------------------------------------------------------\n");
		fprintf(fil, "void %s::Set(const std::string& x)\n", classname.c_str());
		fprintf(fil, "{\n");
		fprintf(fil, "\tm_mixed_content = x;\n");
		fprintf(fil, "}\n\n\n");
	}
}


XmlNode ComplexType::Start() const
{
	XmlNode complexContent(m_node, "complexContent");
	if (complexContent)
	{
		XmlNode extension(complexContent, "extension");
		if (extension)
		{
			if (extension.PropertyExists("base"))
			{
				return extension;
			}
		}
	}
	return m_node;
}


void ComplexType::ShowGenerated()
{
	for (std::map<std::string, bool>::iterator it = m_b_generated.begin(); it != m_b_generated.end(); it++)
	{
		std::cout << "generated: " << it -> first << std::endl;
	}
}


