/**	\file Wsdl.cpp
 **	\date  2008-05-26
 **	\author grymse@alhem.net
**/
/*
Copyright (C) 2008-2009  Anders Hedstrom

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
#include "Wsdl.h"
#include <Sockets/XmlNode.h>
#include "WsdlException.h"
#include "Split.h"
#include <iostream>
#include "ComplexType.h"
#include "Application.h"
#include "SimpleType.h"
#include <Sockets/Debug.h>
#include <Sockets/XmlDocument.h>


#define DEB(x) x; fflush(stdout);


Wsdl::Wsdl(Application& appl, XmlDocument& doc) : m_appl(appl), m_doc(doc)
{
	try
	{
		ParseElements(doc);
	}
	catch (const Exception& e)
	{
		std::cout << "ParseElements() failed: " + e.ToString() << std::endl;
		std::cout << e.Stack() << std::endl;
		abort();
	}
}


Wsdl::~Wsdl()
{
	for (std::list<XmlDocument *>::iterator it = m_import.begin(); it != m_import.end(); it++)
	{
		XmlDocument *p = *it;
		delete p;
	}
}


const std::string Wsdl::GetTargetNamespace() const
{
	XmlNode root(m_doc);
	return root.GetProperty("targetNamespace");
}


const std::string Wsdl::GetTargetNamespacePrefix() const
{
	XmlNode root(m_doc);
	return root.GetNsMapRe()[root.GetProperty("targetNamespace")];
}


std::list<std::string> Wsdl::GetPorts() const
{
	std::list<std::string> vec;
	XmlNode n0(m_doc, "service");
	if (n0)
	{
		XmlNode n(n0, "port");
		while (n)
		{
			vec.push_back(n.GetProperty("name"));
			++n;
		}
	}
	return vec;
}


const std::string Wsdl::GetPortname() const
{
	XmlNode n(m_doc, "portType");
	if (n)
		return n.GetProperty("name");
	throw WsdlException("No portType found in .wsdl document");
}


std::list<std::string> Wsdl::GetOperations(const std::string& portname) const
{
	std::list<std::string> vec;
	XmlNode n(m_doc, "portType");
	while (n)
	{
		if (n.GetProperty("name") == portname || portname.empty())
		{
			XmlNode n2(n, "operation");
			while (n2)
			{
				vec.push_back(n2.GetProperty("name"));
				++n2;
			}
		}
		++n;
	}
	return vec;
}


const std::string Wsdl::GetSoapAction(const std::string& portname, const std::string& op) const
{
	XmlNode n = GetBindingOp(portname, op);
	if (n)
	{
		XmlNode n2(n, "operation");
		if (n2)
		{
			return n2.GetProperty("soapAction");
		}
	}
	throw WsdlException("SoapAction not found");
}


const std::string Wsdl::GetMessageForOp(const std::string& portname, const std::string& op, const std::string& type) const
{
	XmlNode n(m_doc, "portType");
	while (n)
	{
		if (n.GetProperty("name") == portname || portname.empty())
		{
			XmlNode n2(n, "operation");
			while (n2)
			{
				if (n2.GetProperty("name") == op)
				{
					XmlNode n3(n2, type);
					if (n3)
					{
						return n3.GetProperty("message");
					}
				}
				++n2;
			}
		}
		++n;
	}
	throw WsdlException(type + " message not found for: " + portname + "/" + op);
}


XmlNode Wsdl::GetPortTypeOp(const std::string& portname, const std::string& op) const
{
	XmlNode n(m_doc, "portType");
	while (n)
	{
		if (n.GetProperty("name") == portname || portname.empty())
		{
			XmlNode n2(n, "operation");
			while (n2)
			{
				if (n2.GetProperty("name") == op)
				{
					return n2;
				}
				++n2;
			}
		}
		++n;
	}
	throw WsdlException("portType/operation not found: " + portname + "/" + op);
}


XmlNode Wsdl::GetBindingOp(const std::string& portname, const std::string& op) const
{
	XmlNode n(m_doc, "binding");
	while (n)
	{
		if (n.GetProperty("name") == portname || portname.empty())
		{
			XmlNode n2(n, "operation");
			while (n2)
			{
				if (n2.GetProperty("name") == op)
				{
					return n2;
				}
				++n2;
			}
		}
		++n;
	}
	throw WsdlException("binding/operation not found: " + portname + "/" + op);
}


XmlNode Wsdl::GetMessage(const std::string& msgname) const
{
	XmlNode n(m_doc, "message");
	while (n)
	{
		if (n.GetProperty("name") == Split(msgname).Type)
		{
			return n;
		}
		++n;
	}
	throw WsdlException("Message not found: " + msgname);
}


void Wsdl::ShowMessage(const std::string& msgname)
{
	XmlNode message = GetMessage(msgname);
	XmlNode n(message, "part");
	while (n)
	{
		std::string name;
		std::string type;
		std::string element;
		printf("  part/ name: %s", (name = n.GetProperty("name")).c_str());
		try {
			printf(" type: %s", (type = n.GetProperty("type")).c_str());
		} catch (const Exception&) {}
		try {
			printf(" element: %s", (element = n.GetProperty("element")).c_str());
		} catch (const Exception&) {}
		printf("\n");

		if (!element.empty())
		{
			XmlNode el = GetElement(element);
			printf("  {\n");
			ShowElement(el, "    ");
			printf("  }\n");
		}
		++n;
	}
}


void Wsdl::ShowElement(const XmlNode& element_node, const std::string& level)
{
	std::string name;
	std::string ref;
	try {
		name = element_node.GetProperty("name");
	} catch (const Exception&) {}
	try {
		ref = element_node.GetProperty("ref");
	} catch (const Exception&) {}
	XmlNode complex(element_node, "complexType");
	if (complex)
	{
		printf("%s{\n", level.c_str());
		ShowComplexType(complex, level + "  ");
		printf("%s}\n", level.c_str());
		return;
	}
	else
	if (!ref.empty())
	{
		XmlNode n = GetElement( ref );
		if (n)
		{
			printf("%s{\n", level.c_str());
			ShowElement( n, level + "  " );
			printf("%s}\n", level.c_str());
		}
		return;
	}
	std::string type = element_node.GetProperty("type");
	std::string min = "1";
	std::string max = "1";
	try {
		min = element_node.GetProperty("minOccurs");
	} catch (const Exception&) {}
	try {
		max = element_node.GetProperty("maxOccurs");
	} catch (const Exception&) {}
	if (!type.empty())
	{
		std::string tmp = level;
		tmp += type;
		while (tmp.size() < 24 + level.size())
			tmp += " ";
		tmp += " \"" + name + "\"";
		while (tmp.size() < 40 + level.size())
			tmp += " ";
		printf("%s %s .. %s\n", tmp.c_str(), min.c_str(), max.c_str());
		try
		{
			std::string href = element_node.GetNsMap()[Split(type).NS];
			XmlNode n = FindComplexType(href, type);
			if (n)
			{
				printf("%s{\n", level.c_str());
				ShowComplexType(n, level + "  ");
				printf("%s}\n", level.c_str());
			}
		}
		catch (const Exception& ) {}
		return;
	}
	throw WsdlException("no complexType found for element: " + element_node.GetProperty("name"));
}


void Wsdl::ShowComplexType(const XmlNode& complex, const std::string& level)
{
	XmlNode n(complex, "sequence");
	if (n)
	{
		XmlNode n2(n, "element");
		while (n2)
		{
			ShowElement( n2, level );
			++n2;
		}
	}
}


XmlNode Wsdl::GetSchema(const std::string& target_ns) const
{
	XmlNode root(m_doc);
	if (root.GetNodeName() == "schema" &&
		root.GetProperty("targetNamespace") == target_ns)
		return root;
	XmlNode n0(m_doc, "types");
	if (n0)
	{
		XmlNode n(n0, "schema");
		while (n)
		{
			if (n.GetProperty("targetNamespace") == target_ns)
			{
				return n;
			}
			++n;
		}
	}
	throw WsdlException("schema for target namespace not found: " + target_ns);
}


XmlNode Wsdl::GetElement(const std::string& name) const
{
	std::string ns_prefix = Split(name).NS;
	XmlNode root(m_doc);
	std::string ns_href = root.GetNsMap()[ns_prefix];
	XmlNode n = GetSchema(ns_href);
	if (n)
	{
		return GetElement(n, name);
	}
	throw WsdlException("element not found: " + name);
}


XmlNode Wsdl::FindComplexType(const std::string& ns_href, const std::string& type) const
{
	std::map<std::string, std::map<std::string, XmlNode> >::const_iterator it = m_complexTypeMap.find(ns_href);
	if (it == m_complexTypeMap.end())
		throw WsdlException("complexType namespace not found: " + ns_href + "  type: " + type);
	const std::map<std::string, XmlNode>& ref = it -> second;
	std::map<std::string, XmlNode>::const_iterator it2 = ref.find(Split(type).Type);
	if (it2 == ref.end())
		throw WsdlException("Can't find complexType: " + type);
	return it2 -> second;
}


XmlNode Wsdl::FindSimpleType(const std::string& ns_href, const std::string& type) const
{
	std::map<std::string, std::map<std::string, XmlNode> >::const_iterator it = m_simpleTypeMap.find(ns_href);
	if (it == m_simpleTypeMap.end())
		throw WsdlException("simpleType namespace not found: " + ns_href);
	const std::map<std::string, XmlNode>& ref = it -> second;
	std::map<std::string, XmlNode>::const_iterator it2 = ref.find(Split(type).Type);
	if (it2 == ref.end())
		throw WsdlException("Can't find simpleType: " + type);
	return it2 -> second;
}


XmlNode Wsdl::GetElement(const XmlNode& node, const std::string& name) const
{
	{
		XmlNode n(node, "import");
		while (n)
		{
			std::string ns = n.GetProperty("namespace");
			try
			{
				XmlNode ns_node = GetSchema(ns);
				return GetElement(ns_node, name);
			}
			catch (const WsdlException&)
			{
			}
			++n;
		}
	}
	{
		XmlNode n(node, "element");
		while (n)
		{
			if (n.GetProperty("name") == Split(name).Type)
			{
				return n;
			}
			++n;
		}
	}
	throw WsdlException("element not found: " + name);
}


std::list<std::pair<std::string, XmlNode> > Wsdl::GetMessageParameters(const std::string& msgname, std::string& path) const
{
	Debug deb("GetMessageParameters msgname " + msgname + " path " + path);
	std::list<std::pair<std::string, XmlNode> > vec;
	XmlNode message = GetMessage(msgname);
	XmlNode n(message, "part");
	std::string element;

	// part.element --> element
	// part.type --> complex/simple/builtin
	// element.ref --> element
	// element.type --> complex/simple/builtin

	path = "";

	while (n)
	{
		std::string name = n.GetProperty("name");
		if (n.PropertyExists("type"))
		{
			std::string type;
			vec.push_back(std::pair<std::string, XmlNode>(name, n));
		}
		if (n.PropertyExists("element"))
		{
			element = n.GetProperty("element");
		}
		++n;
	}

	XmlNode el(m_doc);

	// check 
	if (vec.size() == 1)
	{
		std::list<std::pair<std::string, XmlNode> >::iterator it = vec.begin();
		XmlNode& ref = it -> second;
		std::string type = ref.GetProperty("type");
		try
		{
			std::string href = ref.GetNsMap()[Split(type).NS];
			XmlNode n = FindComplexType(href, type);
			vec.erase(it);
			el = n;
		} catch (const Exception&) {}
	}

	if (!element.empty()) // go deeper
	{
		el = GetElement(element);
	}
	if (el)
	{
		while (el)
		{
DEB(			printf("/// %s  <%s>\n", el.GetProperty("name").c_str(), el.GetNodeName().c_str());)
			if (el.PropertyExists("name"))
			{
				if (!path.empty())
					path += "/";
				path += el.GetProperty("name");
			}
			if (el.GetNodeName() == "element")
			{
				if (el.PropertyExists("ref"))
				{
					std::string ref = el.GetProperty("ref");
					el = GetElement( ref );
					continue;
				}
				if (el.PropertyExists("type"))
				{
					std::string type = el.GetProperty("type");
					std::string href = el.GetNsMap()[Split(type).NS];
					el = FindComplexType(href, type);
					continue;
				}
				XmlNode n(el, "complexType");
				if (n)
				{
					ComplexType cc(*this, n);
					std::string ref;
					if (cc.NumberOfMembers() == 1 && cc.GetRef(ref))
					{
						el = GetElement( ref );
						continue;
					}
					vec = cc.GetMembers();
				}
				break;
			}
			else
			if (el.GetNodeName() == "complexType")
			{
				ComplexType cc(*this, el);
				std::string ref;
				if (cc.NumberOfMembers() == 1 && cc.GetRef(ref))
				{
					el = GetElement( ref );
					continue;
				}
				vec = cc.GetMembers();
				break;
			}
			else
			{
				throw WsdlException("Handle type <" + el.GetNodeName() + ">");
			}
		}
	}
	return vec;
}


void Wsdl::ParseElements(XmlDocument& doc, const std::string& targetNs)
{
	XmlNode root(doc);
	// schema?
	if (root.GetNodeName() == "schema")
	{
		XmlNode n(root, "import");
		while (n)
		{
			std::string ns = n.PropertyExists("namespace") ? n.GetProperty("namespace") : "";
			std::string loc = n.PropertyExists("schemaLocation") ? n.GetProperty("schemaLocation") : "";
			// namespace, schemaLocation
			std::cout << "Import " << ns << " schemaLocation " << loc << std::endl;
			try
			{
				Utility::Path current = Utility::CurrentDirectory();
				Utility::Path path(loc);
				if (!path.GetPath().empty())
				{
					Utility::ChangeDirectory( path.GetPath() );
				}
				XmlDocument *doc = new XmlDocument( path.GetFilename() );
				m_import.push_back( doc );
				ParseElements( *doc, ns );
				if (!path.GetPath().empty())
				{
					Utility::ChangeDirectory( current );
				}
			}
			catch (const Exception& e)
			{
			}
			++n;
		}
		std::string ns_href = root.GetProperty("targetNamespace");
		if (ns_href == targetNs || targetNs.empty())
		{
			std::string ns_prefix = root.GetNsMapRe()[ns_href];
			ParseTypes(root, ns_href);
			ParseElements(root, ns_href);
		}
	}
	// wsdl schema(s)
	{
		XmlNode n(doc, "types");
		if (n)
		{
			XmlNode n2(n, "schema");
			while (n2)
			{
				XmlNode n3(n2, "import");
				while (n3)
				{
					std::string ns = n3.PropertyExists("namespace") ? n3.GetProperty("namespace") : "";
					std::string loc = n3.PropertyExists("schemaLocation") ? n3.GetProperty("schemaLocation") : "";
					// namespace, schemaLocation
					std::cout << "Import " << ns << " schemaLocation " << loc << std::endl;
					try
					{
						Utility::Path current = Utility::CurrentDirectory();
						Utility::Path path(loc);
						if (!path.GetPath().empty())
						{
							Utility::ChangeDirectory( path.GetPath() );
						}
						XmlDocument *doc = new XmlDocument( path.GetFilename() );
						m_import.push_back( doc );
						ParseElements( *doc, ns );
						if (!path.GetPath().empty())
						{
							Utility::ChangeDirectory( current );
						}
					}
					catch (const Exception& e)
					{
					}
					++n3;
				}
				std::string ns_href = n2.GetProperty("targetNamespace");
				if (ns_href == targetNs || targetNs.empty())
				{
					std::string ns_prefix = root.GetNsMapRe()[ns_href];
					ParseTypes(n2, ns_href);
					ParseElements(n2, ns_href);
				}
				//
				++n2;
			}
		}
	}
}


void Wsdl::ParseTypes(const XmlNode& node, const std::string& ns_href)
{
	{
		XmlNode n(node, "complexType");
		while (n)
		{
			std::string name;
			name = n.GetProperty("name");
			m_complexType[ns_href].push_back(name);
			m_complexTypeMap[ns_href][name] = ParseComplexType(n, ns_href, name);
			++n;
		}
	}

	{
		XmlNode n(node, "simpleType");
		while (n)
		{
			std::string name;
			name = n.GetProperty("name");
			m_simpleType[ns_href].push_back(name);
			m_simpleTypeMap[ns_href][name] = n;
			++n;
		}
	}

}


void Wsdl::ParseElements(const XmlNode& node, const std::string& ns_href, const std::string& parent_class)
{
	XmlNode n(node, "element");
	while (n)
	{
		std::string name;
		if (n.PropertyExists("name"))
			name = n.GetProperty("name");
		std::string type;
		if (n.PropertyExists("type"))
		{
			type = n.GetProperty("type");
		}
		else
		{
			type = name + "_t";
			// test complex- / simpleType
			XmlNode complex(n, "complexType");
			if (complex)
			{
				m_complexType[ns_href].push_back(type);
				m_complexTypeMap[ns_href][type] = ParseComplexType(complex, ns_href, type);
				m_parent[ns_href][type] = parent_class;
			}
			XmlNode simple(n, "simpleType");
			if (simple)
			{
				m_simpleType[ns_href].push_back(type);
				m_simpleTypeMap[ns_href][type] = simple;
				m_parent[ns_href][type] = parent_class;
			}
		}
		++n;
	}
}


XmlNode Wsdl::ParseComplexType(const XmlNode& node, const std::string& ns_href, const std::string& name)
{
	XmlNode start(node);

	XmlNode complexContent(node, "complexContent");
	std::string extends;
	if (complexContent)
	{
		XmlNode extension(complexContent, "extension");
		if (extension)
		{
			if (extension.PropertyExists("base"))
			{
				extends = extension.GetProperty("base");
				start = extension;
			}
		}
	}
	ComplexType t(*this, start, name, ns_href);
	if (t.IsSequence() || t.IsAll())
	{
		XmlNode n(start, t.IsSequence() ? "sequence" : "all");
		if (n)
		{
			ParseElements(n, ns_href, name);
		}
	}
	XmlNode n(start, "attribute");
	while (n)
	{
		std::string aname = n.GetProperty("name");
		if (n.PropertyExists("type"))
		{
		}
		else
		{
			XmlNode n2(n, "simpleType");
			if (n2)
			{
				std::string type = aname + "_t";
				m_simpleType[ns_href].push_back(type);
				m_simpleTypeMap[ns_href][type] = n2;
				m_parent[ns_href][type] = name;
			}
		}
		++n;
	}
	//
	return start;
}


void Wsdl::CreateInterface()
{
	XmlNode root(m_doc);

DEB(	Debug deb("Wsdl::CreateInterface");)
	std::string ns = m_appl.Namespace();
	std::string cn = m_appl.Classname();

	std::string h_filename = cn + ".h";
	std::string cpp_filename = cn + ".cpp";

	FILE *fil = m_appl.myfopen(m_appl.ToDir() + "/" + h_filename, "wt");
	AddLicense(fil);
	fprintf(fil, "#ifndef _WSDL2CPP_%s_H\n", cn.c_str());
	fprintf(fil, "#define _WSDL2CPP_%s_H\n", cn.c_str());
	fprintf(fil, "\n");
	fprintf(fil, "#include <string>\n");
	fprintf(fil, "#include <vector>\n");
	fprintf(fil, "#include <list>\n");
	fprintf(fil, "\n");
	// %!
//	fprintf(fil, "namespace Sockets { class XmlNode; }\n");
	fprintf(fil, "class XmlNode;\n");
	fprintf(fil, "\n");

	{
		NSFil mns(fil, ns); // global namespace

		mns.Begin();

		// simpleType
		for (std::map<std::string, std::list<std::string> >::iterator it = m_simpleType.begin(); it != m_simpleType.end(); it++)
		{
			std::string ns_href = it -> first;
			std::list<std::string>& ref = it -> second;
			if (!ref.empty())
			{
				std::string href = ns_href;
				NSFil mns(fil, GetNamespace(ns_href), href);
				for (std::list<std::string>::iterator it = ref.begin(); it != ref.end(); it++)
				{
					std::string name = *it;
					std::string parent = m_parent[ns_href][name];
					if (parent.empty())
					{
						SimpleType t(*this, m_simpleTypeMap[ns_href][name], name, ns_href);
						mns.Begin();
						t.CreateInterface(fil);
					}
				}
				fprintf(fil, "\n");
			}
		}

		// create class definitions
		bool repeat = false;
		bool gen = false;
		do
		{
			repeat = false;
			gen = false;
			for (std::map<std::string, std::list<std::string> >::iterator it = m_complexType.begin(); it != m_complexType.end(); it++)
			{
				std::string ns_href = it -> first;
DEB(				Debug deb(ns_href);)
				std::list<std::string>& ref = it -> second;
				if (!ref.empty())
				{
					std::string href = ns_href;
					NSFil mns(fil, GetNamespace(ns_href), href);
					for (std::list<std::string>::iterator it = ref.begin(); it != ref.end(); it++)
					{
						std::string name = *it;
						std::string parent = m_parent[ns_href][name];
DEB(						Debug deb(name + " (parent " + parent + ")");)
						if (parent.empty())
						{
							ComplexType t(*this, m_complexTypeMap[ns_href][name], name, ns_href);
							if (!t.IsGenerated())
							{
								if (t.OkToGenerate())
								{
									std::string doc = t.Documentation();
									mns.Begin();
									fprintf(fil, "\t// -----------------------------------------------------------------------------\n");
									if (!doc.empty())
									{
										fprintf(fil, "\t/** %s */\n", doc.c_str());
									}
									t.CreateInterface(fil);
									t.SetGenerated();
									gen = true;
								}
								else
								{
									repeat = true;
								}
							}
						}
					}
				}
			}
		}
		while (repeat && gen);

		//
		fprintf(fil, "\n\n");
	}
	fprintf(fil, "#endif // _WSDL2CPP_%s_H\n", cn.c_str());

	// close .h file
	fclose(fil);

}


void Wsdl::CreateImplementation()
{
	XmlNode root(m_doc);

DEB(	Debug deb("Wsdl::CreateImplementation");)
	std::string ns = m_appl.Namespace();
	std::string cn = m_appl.Classname();

	std::string h_filename = cn + ".h";
	std::string cpp_filename = cn + ".cpp";

	FILE *fil = m_appl.myfopen(m_appl.ToDir() + "/" + cpp_filename, "wt");
	AddLicense(fil);
	{
		fprintf(fil, "#include \"%s\"\n", h_filename.c_str());
		fprintf(fil, "#include <Sockets/XmlNode.h>\n");
		fprintf(fil, "#include <Sockets/XmlException.h>\n");
		fprintf(fil, "#include <Sockets/Utility.h>\n");
		fprintf(fil, "\n");

		NSFil mns(fil, ns); // global namespace
		mns.Begin();

		// create class implementations
		for (std::map<std::string, std::list<std::string> >::iterator it = m_complexType.begin(); it != m_complexType.end(); it++)
		{
			std::string ns_href = it -> first;
			std::list<std::string>& ref = it -> second;
			if (!ref.empty())
			{
				std::string href = ns_href;
				NSFil mns(fil, GetNamespace(ns_href), href);
				mns.Begin();
				for (std::list<std::string>::iterator it = ref.begin(); it != ref.end(); it++)
				{
					std::string name = *it;
					ComplexType t(*this, m_complexTypeMap[ns_href][name], name, ns_href);
					t.CreateImplementation(fil);
				}
			}
		}

	}
	fclose(fil);
}


const std::string Wsdl::GetParent(const std::string& ns_href, const std::string& classname) const
{
DEB(	if (ns_href.empty())
	{
		printf("GetParent( %s, %s )\n", ns_href.c_str(), classname.c_str());
		printf("%s\n", Utility::Stack().c_str());
	})
	std::map<std::string, std::map<std::string, std::string> >::const_iterator it1 = m_parent.find(ns_href);
	if (it1 == m_parent.end())
	{
		return "";
	}
	std::map<std::string, std::string>::const_iterator it;
	if ( (it = it1 -> second.find(classname)) != it1 -> second.end())
	{
		return it -> second;
	}
	return "";
}


const std::list<std::string>& Wsdl::ComplexTypes(const std::string& ns_href) const
{
	std::map<std::string, std::list<std::string> >::const_iterator it = m_complexType.find(ns_href);
	if (it != m_complexType.end())
		return it -> second;
	throw WsdlException("no complexTypes in ns: " + ns_href);
}


const std::list<std::string>& Wsdl::SimpleTypes(const std::string& ns_href) const
{
	std::map<std::string, std::list<std::string> >::const_iterator it = m_simpleType.find(ns_href);
	if (it != m_simpleType.end())
		return it -> second;
	throw WsdlException("no simpleTypes in ns: " + ns_href);
}


const std::string Wsdl::NamespaceFromHref(const std::string& href) const
{
	std::string prefix = m_appl.Href2Prefix( href );
	if (!prefix.empty())
		return prefix;
	size_t pos = href.find("://");
	if (pos != std::string::npos)
	{
		std::string hreft = href.substr(pos + 3);
		pos = hreft.find("/");
		std::string host = hreft;
		std::string path;
		if (pos != std::string::npos)
		{
			host = hreft.substr(0, pos);
			path = hreft.substr(pos + 1);
		}
		Parse pa(host, ".:");
		std::string tmp = pa.getword();
		std::string ns;
		std::string ns2;
		bool first = true;
		while (!tmp.empty())
		{
			if (first)
				first = false;
			else
			{
				ns = ns2;
				ns2 = tmp + "_" + ns2;
			}
			tmp = pa.getword();
		}
		for (size_t i = 0; i < path.size(); i++)
			if (isalnum(path[i]))
				ns2 += path[i];
			else
				ns2 += "_";
		return ns2;
	}
	std::string ns;
	for (size_t i = 0; i < href.size(); i++)
		if (!isalnum(href[i]))
			ns += "_";
		else
			ns += href[i];
	return ns;
}


const std::string Wsdl::GetNamespace(const std::string& ns_href) const
{
	return NamespaceFromHref(ns_href);
}


void Wsdl::AddLicense(FILE *fil)
{
	fprintf(fil, "/*\n"
		"Copyright (C) 2008-2009  Anders Hedstrom (grymse@alhem.net)\n"
		"\n"
		"This program is made available under the terms of the GNU GPL.\n"
		"\n"
		"If you would like to use this program in a closed-source application,\n"
		"a separate license agreement is available. For information about \n"
		"the closed-source license agreement for this program, please\n"
		"visit http://www.alhem.net/Sockets/license.html and/or\n"
		"email license@alhem.net.\n"
		"\n"
		"This program is free software; you can redistribute it and/or\n"
		"modify it under the terms of the GNU General Public License\n"
		"as published by the Free Software Foundation; either version 2\n"
		"of the License, or (at your option) any later version.\n"
		"\n"
		"This program is distributed in the hope that it will be useful,\n"
		"but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
		"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
		"GNU General Public License for more details.\n"
		"\n"
		"You should have received a copy of the GNU General Public License\n"
		"along with this program; if not, write to the Free Software\n"
		"Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.\n"
		"*/\n");
	fprintf(fil, "\n");
}


